@@ -4,9 +4,9 @@ var Stream = require('net').Stream;
 var InternalChildProcess = process.binding('child_process').ChildProcess;
 
 
-var spawn = exports.spawn = function (path, args, env) {
+var spawn = exports.spawn = function (path, args, env, customFds) {
   var child = new ChildProcess();
-  child.spawn(path, args, env);
+  child.spawn(path, args, env, customFds);
   return child;
 };
 
@@ -86,7 +86,7 @@ exports.execFile = function (file, args /*, options, callback */) {
 
 
 function ChildProcess () {
-  process.EventEmitter.call(this);
+  EventEmitter.call(this);
 
   var self = this;
 
@@ -99,18 +99,32 @@ function ChildProcess () {
   var stdout = this.stdout = new Stream();
   var stderr = this.stderr = new Stream();
 
-  stderr.onend = stdout.onend = function () {
-    if (gotCHLD && !stdout.readable && !stderr.readable) {
+  var stderrClosed = false;
+  var stdoutClosed = false;
+
+  stderr.addListener('close', function () {
+    stderrClosed = true;
+    if (gotCHLD && (!self.stdout || stdoutClosed)) {
       self.emit('exit', exitCode, termSignal);
     }
-  };
+  });
+
+  stdout.addListener('close', function () {
+    stdoutClosed = true;
+    if (gotCHLD && (!self.stderr || stderrClosed)) {
+      self.emit('exit', exitCode, termSignal);
+    }
+  });
 
   internal.onexit = function (code, signal) {
     gotCHLD = true;
     exitCode = code;
     termSignal = signal;
-    stdin.end();
-    if (!stdout.readable && !stderr.readable) {
+    if (self.stdin) {
+      self.stdin.end();
+    }
+    if (   (!self.stdout || !self.stdout.readable)
+        && (!self.stderr || !self.stderr.readable)) {
       self.emit('exit', exitCode, termSignal);
     }
   };
@@ -125,7 +139,7 @@ ChildProcess.prototype.kill = function (sig) {
 };
 
 
-ChildProcess.prototype.spawn = function (path, args, env) {
+ChildProcess.prototype.spawn = function (path, args, env, customFds) {
   args = args || [];
   env = env || process.env;
   var envPairs = [];
@@ -135,20 +149,36 @@ ChildProcess.prototype.spawn = function (path, args, env) {
     envPairs.push(key + "=" + env[key]);
   }
 
-  var fds = this._internal.spawn(path, args, envPairs);
+  customFds = customFds || [-1, -1, -1];
+  var fds = this.fds = this._internal.spawn(path, args, envPairs, customFds);
 
-  this.stdin.open(fds[0]);
-  this.stdin.writable = true;
-  this.stdin.readable = false;
+  if (customFds[0] === -1 || customFds[0] === undefined) {
+    this.stdin.open(fds[0]);
+    this.stdin.writable = true;
+    this.stdin.readable = false;
+  }
+  else {
+    this.stdin = null;
+  }
 
-  this.stdout.open(fds[1]);
-  this.stdout.writable = false;
-  this.stdout.readable = true;
-  this.stdout.resume();
+  if (customFds[1] === -1 || customFds[1] === undefined) {
+    this.stdout.open(fds[1]);
+    this.stdout.writable = false;
+    this.stdout.readable = true;
+    this.stdout.resume();
+  }
+  else {
+    this.stdout = null;
+  }
 
-  this.stderr.open(fds[2]);
-  this.stderr.writable = false;
-  this.stderr.readable = true;
-  this.stderr.resume();
+  if (customFds[2] === -1 || customFds[2] === undefined) {
+    this.stderr.open(fds[2]);
+    this.stderr.writable = false;
+    this.stderr.readable = true;
+    this.stderr.resume();
+  }
+  else {
+    this.stderr = null;
+  }
 };
 
