@@ -1,9 +1,13 @@
-var sys = require('sys'),
-    events = require('events');
+var sys = require('sys');
+var events = require('events');
+var Buffer = require('buffer').Buffer;
 
 var binding = process.binding('fs');
 var fs = exports;
 
+var kMinPoolSpace = 128;
+var kPoolSize = 40*1024;
+
 fs.Stats = binding.Stats;
 
 fs.Stats.prototype._checkModeProperty = function (property) {
@@ -38,53 +42,70 @@ fs.Stats.prototype.isSocket = function () {
   return this._checkModeProperty(process.S_IFSOCK);
 };
 
-
-function readAll (fd, pos, content, encoding, callback) {
-  binding.read(fd, 4*1024, pos, encoding, function (err, chunk, bytesRead) {
-    if (err) {
-      if (callback) callback(err);
-    } else if (chunk) {
-      content += chunk;
-      pos += bytesRead;
-      readAll(fd, pos, content, encoding, callback);
-    } else {
-      binding.close(fd, function (err) {
-        if (callback) callback(err, content);
-      });
-    }
-  });
-}
-
 fs.readFile = function (path, encoding_, callback) {
-  var encoding = typeof(encoding_) == 'string' ? encoding_ : 'utf8';
+  var encoding = typeof(encoding_) == 'string' ? encoding_ : null;
   var callback_ = arguments[arguments.length - 1];
-  var callback = (typeof(callback_) == 'function' ? callback_ : null);
-  binding.open(path, process.O_RDONLY, 0666, function (err, fd) {
-    if (err) {
-      if (callback) callback(err);
-    } else {
-      // leave pos null to allow reads on unseekable devices
-      readAll(fd, null, "", encoding, callback);
-    }
+  var callback = (typeof(callback_) == 'function' ? callback_ : noop);
+  binding.stat(path, function (err, stat) {
+    if (err) { callback(err); return; }
+    binding.open(path, process.O_RDONLY, 0666, function (err, fd) {
+      if (err) { callback(err); return; }
+      var size = stat.size;
+      var buffer = new Buffer(size);
+      var offset = 0;
+      function doRead() {
+        if (size < 1) {
+          callback(null, buffer);
+          return;
+        }
+        // position is offset or null so we can read files on unseekable mediums
+        binding.read(fd, buffer, offset, size - offset, offset || null, function (err, amount) {
+          if (err) {
+            callback(err);
+            binding.close(fd);
+            return;
+          }
+          if (amount + offset < size) {
+            offset += amount;
+            doRead();
+            return;
+          }
+          binding.close(fd);
+          if (encoding) {
+            try {
+              var str = buffer.toString(encoding);
+            } catch (err) {
+              callback(err);
+              return;
+            }
+            callback(null, str);
+          } else {
+            callback(null, buffer);
+          }
+        });
+      }
+      doRead();
+    });
   });
 };
 
 fs.readFileSync = function (path, encoding) {
-  encoding = encoding || "utf8"; // default to utf8
+  var fd = fs.openSync(path, process.O_RDONLY, 0666);
+  var stat = fs.statSync(path);
+  var buffer = new Buffer(stat.size);
+  var nread = 0;
 
-  var fd = binding.open(path, process.O_RDONLY, 0666);
-  var content = '';
-  var pos = null;   // leave null to allow reads on unseekable devices
-  var r;
-
-  while ((r = binding.read(fd, 4*1024, pos, encoding)) && r[0]) {
-    content += r[0];
-    pos += r[1]
+  while (nread < buffer.length) {
+    nread += fs.readSync(fd, buffer, nread, buffer.length - nread, null);
   }
 
-  binding.close(fd);
+  fs.closeSync(fd);
 
-  return content;
+  if (encoding) {
+    return buffer.toString(encoding);
+  } else {
+    return buffer;
+  }
 };
 
 
@@ -128,24 +149,79 @@ fs.openSync = function (path, flags, mode) {
   return binding.open(path, stringToFlags(flags), mode);
 };
 
-fs.read = function (fd, length, position, encoding, callback) {
-  encoding = encoding || "binary";
-  binding.read(fd, length, position, encoding, callback || noop);
+fs.read = function (fd, buffer, offset, length, position, callback) {
+  if (!(buffer instanceof Buffer)) {
+    // legacy string interface (fd, length, position, encoding, callback)
+    var cb = arguments[4],
+        encoding = arguments[3];
+    position = arguments[2];
+    length = arguments[1];
+    buffer = new Buffer(length);
+    offset = 0;
+
+    callback = function(err, bytesRead) {
+      if (!cb) return;
+
+      var str = (bytesRead > 0)
+        ? buffer.toString(encoding, 0, bytesRead)
+        : '';
+
+      (cb)(err, str, bytesRead);
+    };
+  }
+
+  binding.read(fd, buffer, offset, length, position, callback || noop);
 };
 
-fs.readSync = function (fd, length, position, encoding) {
-  encoding = encoding || "binary";
-  return binding.read(fd, length, position, encoding);
+fs.readSync = function (fd, buffer, offset, length, position) {
+  var legacy = false;
+  if (!(buffer instanceof Buffer)) {
+    // legacy string interface (fd, length, position, encoding, callback)
+    legacy = true;
+    encoding = arguments[3];
+    position = arguments[2];
+    length = arguments[1];
+    buffer = new Buffer(length);
+
+    offset = 0;
+  }
+
+  var r = binding.read(fd, buffer, offset, length, position);
+  if (!legacy) {
+    return r;
+  }
+
+  var str = (r > 0)
+      ? buffer.toString(encoding, 0, r)
+      : '';
+  return [str, r];
 };
 
-fs.write = function (fd, data, position, encoding, callback) {
-  encoding = encoding || "binary";
-  binding.write(fd, data, position, encoding, callback || noop);
+fs.write = function (fd, buffer, offset, length, position, callback) {
+  if (!(buffer instanceof Buffer)) {
+    // legacy string interface (fd, data, position, encoding, callback)
+    callback = arguments[4];
+    position = arguments[2];
+
+    buffer = new Buffer(''+arguments[1], arguments[3]);
+    offset = 0;
+    length = buffer.length;
+  }
+
+  binding.write(fd, buffer, offset, length, position, callback || noop);
 };
 
-fs.writeSync = function (fd, data, position, encoding) {
-  encoding = encoding || "binary";
-  return binding.write(fd, data, position, encoding);
+fs.writeSync = function (fd, buffer, offset, length, position) {
+  if (!(buffer instanceof Buffer)) {
+    // legacy string interface (fd, data, position, encoding)
+    position = arguments[2];
+
+    buffer = new Buffer(''+arguments[1], arguments[3]);
+    offset = 0;
+    length = buffer.length;
+  }
+
+  binding.write(fd, buffer, offset, length, position);
 };
 
 fs.rename = function (oldPath, newPath, callback) {
@@ -172,6 +248,22 @@ fs.rmdirSync = function (path) {
   return binding.rmdir(path);
 };
 
+fs.fdatasync = function (fd, callback) {
+  binding.fdatasync(fd, callback || noop);
+};
+
+fs.fdatasyncSync = function (fd) {
+  return binding.fdatasync(fd);
+};
+
+fs.fsync = function (fd, callback) {
+  binding.fsync(fd, callback || noop);
+};
+
+fs.fsyncSync = function (fd) {
+  return binding.fsync(fd);
+};
+
 fs.mkdir = function (path, mode, callback) {
   binding.mkdir(path, mode, callback || noop);
 };
@@ -196,6 +288,10 @@ fs.readdirSync = function (path) {
   return binding.readdir(path);
 };
 
+fs.fstat = function (fd, callback) {
+  binding.fstat(fd, callback || noop);
+};
+
 fs.lstat = function (path, callback) {
   binding.lstat(path, callback || noop);
 };
@@ -204,6 +300,10 @@ fs.stat = function (path, callback) {
   binding.stat(path, callback || noop);
 };
 
+fs.fstatSync = function (fd) {
+  return binding.fstat(fd);
+};
+
 fs.lstatSync = function (path) {
   return binding.lstat(path);
 };
@@ -252,6 +352,14 @@ fs.chmodSync = function (path, mode) {
   return binding.chmod(path, mode);
 };
 
+fs.chown = function(path, uid, gid, callback) {
+  binding.chown(path, uid, gid, callback || noop);
+};
+
+fs.chownSync = function(path, uid, gid) {
+  return binding.chown(path, uid, gid);
+};
+
 function writeAll (fd, data, encoding, callback) {
   fs.write(fd, data, 0, encoding, function (writeErr, written) {
     if (writeErr) {
@@ -343,8 +451,8 @@ fs.unwatchFile = function (filename) {
 // Realpath
 
 var path = require('path');
-var normalize = path.normalize
-    normalizeArray = path.normalizeArray;
+var normalize = path.normalize;
+var normalizeArray = path.normalizeArray;
 
 fs.realpathSync = function (path) {
   var seen_links = {}, knownHards = {}, buf, i = 0, part, x, stats;
@@ -471,8 +579,16 @@ fs.realpath = function (path, callback) {
     }
   }
   next();
+};
+
+var pool;
+function allocNewPool () {
+  pool = new Buffer(kPoolSize);
+  pool.used = 0;
 }
 
+
+
 fs.createReadStream = function(path, options) {
   return new ReadStream(path, options);
 };
@@ -486,7 +602,6 @@ var ReadStream = fs.ReadStream = function(path, options) {
   this.paused = false;
 
   this.flags = 'r';
-  this.encoding = 'binary';
   this.mode = 0666;
   this.bufferSize = 4 * 1024;
 
@@ -520,19 +635,40 @@ sys.inherits(ReadStream, events.EventEmitter);
 
 fs.FileReadStream = fs.ReadStream; // support the legacy name
 
-ReadStream.prototype.setEncoding = function(encoding) {
-  this.encoding = encoding;
+ReadStream.prototype.setEncoding = function (encoding) {
+  var Utf8Decoder = require("utf8decoder").Utf8Decoder; // lazy load
+  var self = this;
+  this._encoding = encoding.toLowerCase();
+  if (this._encoding == 'utf-8' || this._encoding == 'utf8') {
+    this._decoder = new Utf8Decoder();
+    this._decoder.onString = function(str) {
+      self.emit('data', str);
+    };
+  } else if (this._decoder) {
+    delete this._decoder;
+  }
 };
 
+
 ReadStream.prototype._read = function () {
   var self = this;
   if (!self.readable || self.paused) return;
 
-  fs.read(self.fd,
-          self.bufferSize,
-          undefined,
-          self.encoding,
-          function(err, data, bytesRead) {
+  if (!pool || pool.length - pool.used < kMinPoolSpace) {
+    // discard the old pool. Can't add to the free list because
+    // users might have refernces to slices on it.
+    pool = null;
+    allocNewPool();
+  }
+
+  // Grab another reference to the pool in the case that while we're in the
+  // thread pool another read() finishes up the pool, and allocates a new
+  // one.
+  var thisPool = pool;
+  var toRead = Math.min(pool.length - pool.used, this.bufferSize);
+  var start = pool.used;
+
+  function afterRead (err, bytesRead) {
     if (err) {
       self.emit('error', err);
       self.readable = false;
@@ -545,20 +681,39 @@ ReadStream.prototype._read = function () {
       return;
     }
 
+    var b = thisPool.slice(start, start+bytesRead);
+
+    // Possible optimizition here?
+    // Reclaim some bytes if bytesRead < toRead?
+    // Would need to ensure that pool === thisPool.
+
     // do not emit events if the stream is paused
     if (self.paused) {
-      self.buffer = data;
+      self.buffer = b;
       return;
     }
 
     // do not emit events anymore after we declared the stream unreadable
-    if (!self.readable) {
-      return;
-    }
+    if (!self.readable) return;
 
-    self.emit('data', data);
+    self._emitData(b);
     self._read();
-  });
+  }
+
+  fs.read(self.fd, pool, pool.used, toRead, undefined, afterRead);
+  pool.used += toRead;
+};
+
+
+ReadStream.prototype._emitData = function (d) {
+  if (!this._encoding) {
+    this.emit('data', d);
+  } else if (this._decoder) {
+    this._decoder.write(d);
+  } else {
+    var string = d.toString(this._encoding, 0, d.length);
+    this.emit('data', string);
+  }
 };
 
 
@@ -570,7 +725,7 @@ ReadStream.prototype.forceClose = function (cb) {
     sys.error(readStreamForceCloseWarning);
   }
   return this.destroy(cb);
-}
+};
 
 
 ReadStream.prototype.destroy = function (cb) {
@@ -580,16 +735,12 @@ ReadStream.prototype.destroy = function (cb) {
   function close() {
     fs.close(self.fd, function(err) {
       if (err) {
-        if (cb) {
-          cb(err);
-        }
+        if (cb) cb(err);
         self.emit('error', err);
         return;
       }
 
-      if (cb) {
-        cb(null);
-      }
+      if (cb) cb(null);
       self.emit('close');
     });
   }
@@ -611,7 +762,7 @@ ReadStream.prototype.resume = function() {
   this.paused = false;
 
   if (this.buffer) {
-    this.emit('data', this.buffer);
+    this._emitData(this.buffer);
     this.buffer = null;
   }
 
@@ -713,14 +864,25 @@ WriteStream.prototype.flush = function () {
 };
 
 
-WriteStream.prototype.write = function(data, encoding) {
+WriteStream.prototype.write = function (data) {
   if (!this.writeable) {
     throw new Error('stream not writeable');
   }
 
-  // TODO handle Buffer
+  var cb;
+  if (typeof(arguments[arguments.length-1]) == 'function') {
+    cb = arguments[arguments.length-1];
+  }
+
+  if (data instanceof Buffer) {
+    this._queue.push([fs.write, data, 0, data.length, null, cb]);
+  } else {
+    var encoding = 'utf8';
+    if (typeof(arguments[1]) == 'string') encoding = arguments[1];
+    this._queue.push([fs.write, data, undefined, encoding, cb]);
+  }
+
 
-  this._queue.push([fs.write, data, undefined, encoding || 'utf8', null]);
   this.flush();
 
   return false;
@@ -753,7 +915,7 @@ WriteStream.prototype.forceClose = function (cb) {
     sys.error(writeStreamForceCloseWarning);
   }
   return this.destroy(cb);
-}
+};
 
 
 WriteStream.prototype.forceClose = function (cb) {
