@@ -38,7 +38,20 @@ exports.normalize = function (path, keepBlanks) {
 };
 
 exports.dirname = function (path) {
-  return path && path.substr(0, path.lastIndexOf("/")) || ".";
+  // Can't strip trailing slashes since module.js incorrectly thinks
+  // dirname('/a/b/') should yield '/a/b' instead of '/a'.
+  // if (path.length > 1 && '/' === path[path.length-1]) {
+  //  path = path.replace(/\/+$/, '');
+  // }
+  var lastSlash = path.lastIndexOf('/');
+  switch (lastSlash) {
+    case -1:
+      return '.';
+    case 0:
+      return '/';
+    default:
+      return path.substring(0, lastSlash);
+  }
 };
 
 exports.filename = function () {
@@ -53,8 +66,12 @@ exports.basename = function (path, ext) {
 };
 
 exports.extname = function (path) {
-  var index = path.lastIndexOf('.');
-  return index < 0 ? '' : path.substring(index);
+  var dot = path.lastIndexOf('.'),
+      slash = path.lastIndexOf('/');
+  // The last dot must be in the last path component, and it (the last dot) must
+  // not start the last path component (i.e. be a dot that signifies a hidden
+  // file in UNIX).
+  return dot <= slash + 1 ? '' : path.substring(dot);
 };
 
 exports.exists = function (path, callback) {
