@@ -13,21 +13,105 @@
 // repl.start("node > ").scope.foo = "stdin is fun";  // expose foo to repl scope
 
 var sys = require('sys');
+var evalcx = process.binding('evals').Script.runInNewContext;
+var path = require("path");
+var rl = require('readline');
+var scope;
+
+function cwdRequire (id) {
+  if (id.match(/^\.\.\//) || id.match(/^\.\//)) {
+    id = path.join(process.cwd(), id);
+  }
+  return require(id);
+}
+
+function setScope (self) {
+  scope = {};
+  for (var i in global) scope[i] = global[i];
+  scope.module = module;
+  scope.require = cwdRequire;
+}
+
 
 // Can overridden with custom print functions, such as `probe` or `eyes.js`
 exports.writer = sys.inspect;
 
 function REPLServer(prompt, stream) {
   var self = this;
-  
-  self.scope = {};
+  if (!scope) setScope();
+  self.scope = scope;
   self.buffered_cmd = '';
-  self.prompt = prompt || "node> ";
+
   self.stream = stream || process.openStdin();
-  self.stream.setEncoding('utf8');
+  self.prompt = prompt || "node> ";
+
+  var rli = self.rli = rl.createInterface(self.stream);
+  rli.setPrompt(self.prompt);
+
   self.stream.addListener("data", function (chunk) {
-    self.readline.call(self, chunk);
+    rli.write(chunk);
   });
+
+  rli.addListener('line', function (cmd) {
+    cmd = trimWhitespace(cmd);
+
+    var flushed = true;
+
+    // Check to see if a REPL keyword was used. If it returns true,
+    // display next prompt and return.
+    if (self.parseREPLKeyword(cmd) === true) return;
+
+    // The catchall for errors
+    try {
+      self.buffered_cmd += cmd;
+      // This try is for determining if the command is complete, or should
+      // continue onto the next line.
+      try {
+        // Use evalcx to supply the global scope
+        var ret = evalcx(self.buffered_cmd, scope, "repl");
+        if (ret !== undefined) {
+          scope._ = ret;
+          flushed = self.stream.write(exports.writer(ret) + "\n");
+        }
+
+        self.buffered_cmd = '';
+      } catch (e) {
+        // instanceof doesn't work across context switches.
+        if (!(e && e.constructor && e.constructor.name === "SyntaxError")) {
+          throw e;
+        }
+      }
+    } catch (e) {
+      // On error: Print the error and clear the buffer
+      if (e.stack) {
+        flushed = self.stream.write(e.stack + "\n");
+      } else {
+        flushed = self.stream.write(e.toString() + "\n");
+      }
+      self.buffered_cmd = '';
+    }
+
+    // need to make sure the buffer is flushed before displaying the prompt
+    // again. This is really ugly. Need to have callbacks from
+    // net.Stream.write()
+    if (flushed) {
+      self.displayPrompt();
+    } else {
+      self.displayPromptOnDrain = true;
+    }
+  });
+
+  self.stream.addListener('drain', function () {
+    if (self.displayPromptOnDrain) {
+      self.displayPrompt();
+      self.displayPromptOnDrain = false;
+    }
+  });
+
+  rli.addListener('close', function () {
+    self.stream.destroy();
+  });
+
   self.displayPrompt();
 }
 exports.REPLServer = REPLServer;
@@ -39,54 +123,12 @@ exports.start = function (prompt, source) {
 };
 
 REPLServer.prototype.displayPrompt = function () {
-  var self = this;
-  self.stream.write(self.buffered_cmd.length ? '...   ' : self.prompt);
+  this.rli.setPrompt(this.buffered_cmd.length ? '...   ' : this.prompt);
+  this.rli.prompt();
 };
 
 // read a line from the stream, then eval it
 REPLServer.prototype.readline = function (cmd) {
-  var self = this;
-
-  cmd = self.trimWhitespace(cmd);
-  
-  // Check to see if a REPL keyword was used. If it returns true,
-  // display next prompt and return.
-  if (self.parseREPLKeyword(cmd) === true) {
-    return;
-  }
-  
-  // The catchall for errors
-  try {
-    self.buffered_cmd += cmd;
-    // This try is for determining if the command is complete, or should
-    // continue onto the next line.
-    try {
-      self.buffered_cmd = self.convertToScope(self.buffered_cmd);
-      
-      // Scope the readline with self.scope to provide "local" vars and make Douglas Crockford cry
-      with (self.scope) {
-        var ret = eval(self.buffered_cmd);
-        if (ret !== undefined) {
-          self.scope['_'] = ret;
-          self.stream.write(exports.writer(ret) + "\n");
-        }
-      }
-        
-      self.buffered_cmd = '';
-    } catch (e) {
-      if (!(e instanceof SyntaxError)) throw e;
-    }
-  } catch (e) {
-    // On error: Print the error and clear the buffer
-    if (e.stack) {
-      self.stream.write(e.stack + "\n");
-    } else {
-      self.stream.write(e.toString() + "\n");
-    }
-    self.buffered_cmd = '';
-  }
-  
-  self.displayPrompt();
 };
 
 /**
@@ -107,11 +149,11 @@ REPLServer.prototype.parseREPLKeyword = function (cmd) {
   case ".clear":
     self.stream.write("Clearing Scope...\n");
     self.buffered_cmd = '';
-    self.scope = {};
+    setScope();
     self.displayPrompt();
     return true;
   case ".exit":
-    self.stream.end();
+    self.stream.destroy();
     return true;
   case ".help":
     self.stream.write(".break\tSometimes you get stuck in a place you can't get out... This will get you out.\n");
@@ -124,20 +166,14 @@ REPLServer.prototype.parseREPLKeyword = function (cmd) {
   return false;
 };
 
-/**
- * Trims Whitespace from a line.
- * 
- * @param {String} cmd The string to trim the whitespace from
- * @returns {String} The trimmed string 
- */
-REPLServer.prototype.trimWhitespace = function (cmd) {
+function trimWhitespace (cmd) {
   var trimmer = /^\s*(.+)\s*$/m, 
     matches = trimmer.exec(cmd);
 
   if (matches && matches.length === 2) {
     return matches[1];
   }
-};
+}
 
 /**
  * Converts commands that use var and function <name>() to use the
