@@ -1,8 +1,9 @@
 require("../common");
 var path = require('path');
+var Buffer = require('buffer').Buffer;
 var fs = require('fs');
 var fn = path.join(fixturesDir, "write.txt");
-var expected = "hello";
+var expected = "ümlaut.";
 var found;
 
 fs.open(fn, 'w', 0644, function (err, fd) {
@@ -11,7 +12,7 @@ fs.open(fn, 'w', 0644, function (err, fd) {
   fs.write(fd, expected, 0, "utf8", function (err, written) {
     puts('write done');
     if (err) throw err;
-    assert.equal(expected.length, written);
+    assert.equal(Buffer.byteLength(expected), written);
     fs.closeSync(fd);
     found = fs.readFileSync(fn, 'utf8');
     puts('expected: ' + expected.toJSON());
