@@ -1,5 +1,5 @@
 require("../common");
-tcp = require("tcp");
+net = require("net");
 http = require("http");
 url = require("url");
 
@@ -14,16 +14,16 @@ var s = http.createServer(function (req, res) {
 
   res.writeHead(200, {"Content-Type": "text/plain"});
   res.write("Hello World");
-  res.close();
+  res.end();
 
   if (++nrequests_completed == nrequests_expected) s.close();
 });
 s.listen(PORT);
 
-var c = tcp.createConnection(PORT);
+var c = net.createConnection(PORT);
 c.addListener("connect", function () {
   c.write("GET /hello?foo=%99bar HTTP/1.1\r\n\r\n");
-  c.close();
+  c.end();
 });
 
 //  TODO add more!
