@@ -6,12 +6,12 @@ var PROXY_PORT = PORT;
 var BACKEND_PORT = PORT+1;
 
 var backend = http.createServer(function (req, res) {
-  // debug("backend");
+  debug("backend request");
   res.writeHead(200, {"content-type": "text/plain"});
   res.write("hello world\n");
-  res.close();
+  res.end();
 });
-// debug("listen backend")
+debug("listen backend")
 backend.listen(BACKEND_PORT);
 
 var proxy_client = http.createClient(BACKEND_PORT);
@@ -24,32 +24,41 @@ var proxy = http.createServer(function (req, res) {
       res.write(chunk);
     });
     proxy_res.addListener("end", function() {
-      res.close();
-      // debug("proxy res");
+      res.end();
+      debug("proxy res");
     });
   });
-  proxy_req.close();
+  proxy_req.end();
 });
-// debug("listen proxy")
+debug("listen proxy")
 proxy.listen(PROXY_PORT);
 
 var body = "";
 
-var client = http.createClient(PROXY_PORT);
-var req = client.request("/test");
-// debug("client req")
-req.addListener('response', function (res) {
-  // debug("got res");
-  assert.equal(200, res.statusCode);
-  res.setBodyEncoding("utf8");
-  res.addListener('data', function (chunk) { body += chunk; });
-  res.addListener('end', function () {
-    proxy.close();
-    backend.close();
-    // debug("closed both");
+nlistening = 0;
+function startReq () {
+  nlistening++;
+  if (nlistening < 2) return;
+
+  var client = http.createClient(PROXY_PORT);
+  var req = client.request("/test");
+  debug("client req")
+  req.addListener('response', function (res) {
+    debug("got res");
+    assert.equal(200, res.statusCode);
+    res.setBodyEncoding("utf8");
+    res.addListener('data', function (chunk) { body += chunk; });
+    res.addListener('end', function () {
+      proxy.close();
+      backend.close();
+       debug("closed both");
+    });
   });
-});
-req.close();
+  req.end();
+}
+
+proxy.addListener('listening', startReq);
+backend.addListener('listening', startReq);
 
 process.addListener("exit", function () {
   assert.equal(body, "hello world\n");
